@@ -1,7 +1,7 @@
 /*
     webalizer - a web server log analysis program
 
-    Copyright (C) 1997-2009  Bradford L. Barrett
+    Copyright (C) 1997-2010  Bradford L. Barrett
 
     This program is free software; you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published by
@@ -68,7 +68,7 @@ DNODEPTR new_dnode(char *);                   /* new DNS node             */
 void     update_entry(char *);                /* update entry/exit        */
 void     update_exit(char *);                 /* page totals              */
 
-u_int64_t hash(char *);                       /* hash function            */
+uint32_t hash(char *);                        /* hash function            */
 
 /* local data */
 
@@ -83,6 +83,10 @@ INODEPTR im_htab[MAXHASH];                    /* ident table (username)   */
 DNODEPTR host_table[MAXHASH];                 /* DNS hash table           */
 #endif  /* USE_DNS */
 
+/* Last node pointers */
+HNODEPTR lm_hnode=NULL;
+HNODEPTR ld_hnode=NULL;
+RNODEPTR l_rnode=NULL;
 
 /*********************************************/
 /* DEL_HTABS - clear out our hash tables     */
@@ -153,19 +157,23 @@ int put_hnode( char      *str,  /* Hostname  */
                HNODEPTR  *htab)  /* ptr>next  */
 {
    HNODEPTR cptr,nptr;
+   uint32_t  hval;
 
    /* check if hashed */
-   if ( (cptr = htab[hash(str)]) == NULL)
+   hval=hash(str);
+   if ( (cptr = htab[hval]) == NULL)
    {
       /* not hashed */
       if ( (nptr=new_hnode(str)) != NULL)
       {
+         if (htab==sm_htab) lm_hnode=nptr;
+         else               ld_hnode=nptr;
          nptr->flag  = type;
          nptr->count = count;
          nptr->files = file;
          nptr->xfer  = xfer;
          nptr->next  = NULL;
-         htab[hash(str)] = nptr;
+         htab[hval] = nptr;
          if (type!=OBJ_GRP) (*ctr)++;
 
          if (visit)
@@ -189,7 +197,16 @@ int put_hnode( char      *str,  /* Hostname  */
    }
    else
    {
-      /* hashed */
+      /* hashed (SPEEDUP) */
+      if (htab==sm_htab)
+      {
+         if (lm_hnode!=NULL && strcmp(lm_hnode->string,str)==0) cptr=lm_hnode;
+      }
+      else
+      {
+         if (ld_hnode!=NULL && strcmp(ld_hnode->string,str)==0) cptr=ld_hnode;
+      }
+
       while (cptr != NULL)
       {
          if (strcmp(cptr->string,str)==0)
@@ -215,6 +232,8 @@ int put_hnode( char      *str,  /* Hostname  */
                   cptr->lasturl=find_url(log_rec.url);
                   cptr->tstamp=tstamp;
                }
+               if (htab==sm_htab) lm_hnode=cptr;
+               else               ld_hnode=cptr;
                return 0;
             }
          }
@@ -223,12 +242,14 @@ int put_hnode( char      *str,  /* Hostname  */
       /* not found... */
       if ( (nptr = new_hnode(str)) != NULL)
       {
+         if (htab==sm_htab) lm_hnode=nptr;
+         else               ld_hnode=nptr;
          nptr->flag  = type;
          nptr->count = count;
          nptr->files = file;
          nptr->xfer  = xfer;
-         nptr->next  = htab[hash(str)];
-         htab[hash(str)]=nptr;
+         nptr->next  = htab[hval];
+         htab[hval]=nptr;
          if (type!=OBJ_GRP) (*ctr)++;
 
          if (visit)
@@ -260,6 +281,8 @@ int put_hnode( char      *str,  /* Hostname  */
          /* check if it's a hidden object */
          if ((hide_sites)||(isinlist(hidden_sites,nptr->string)!=NULL))
            nptr->flag=OBJ_HIDE;
+         if (htab==sm_htab) lm_hnode=nptr;
+         else               ld_hnode=nptr;
       }
    }
    return nptr==NULL;
@@ -290,6 +313,8 @@ void	del_hlist(HNODEPTR *htab)
          htab[i]=NULL;
       }
    }
+   lm_hnode=NULL;
+   ld_hnode=NULL;
 }
 
 /*********************************************/
@@ -334,11 +359,13 @@ int put_unode(char *str, int type, u_int64_t count, double xfer,
               u_int64_t *ctr, u_int64_t entry, u_int64_t exit, UNODEPTR *htab)
 {
    UNODEPTR cptr,nptr;
+   uint32_t  hval;
 
    if (str[0]=='-') return 0;
 
+   hval = hash(str);
    /* check if hashed */
-   if ( (cptr = htab[hash(str)]) == NULL)
+   if ( (cptr = htab[hval]) == NULL)
    {
       /* not hashed */
       if ( (nptr=new_unode(str)) != NULL)
@@ -349,7 +376,7 @@ int put_unode(char *str, int type, u_int64_t count, double xfer,
          nptr->next = NULL;
          nptr->entry= entry;
          nptr->exit = exit;
-         htab[hash(str)] = nptr;
+         htab[hval] = nptr;
          if (type!=OBJ_GRP) (*ctr)++;
       }
    }
@@ -376,10 +403,10 @@ int put_unode(char *str, int type, u_int64_t count, double xfer,
          nptr->flag = type;
          nptr->count= count;
          nptr->xfer = xfer;
-         nptr->next = htab[hash(str)];
+         nptr->next = htab[hval];
          nptr->entry= entry;
          nptr->exit = exit;
-         htab[hash(str)]=nptr;
+         htab[hval] = nptr;
          if (type!=OBJ_GRP) (*ctr)++;
       }
    }
@@ -461,11 +488,13 @@ int put_rnode(char *str, int type, u_int64_t count,
               u_int64_t *ctr, RNODEPTR *htab)
 {
    RNODEPTR cptr,nptr;
+   uint32_t  hval;
 
    if (str[0]=='-') strcpy(str,"- (Direct Request)");
 
+   hval = hash(str);
    /* check if hashed */
-   if ( (cptr = htab[hash(str)]) == NULL)
+   if ( (cptr = htab[hval]) == NULL)
    {
       /* not hashed */
       if ( (nptr=new_rnode(str)) != NULL)
@@ -473,13 +502,15 @@ int put_rnode(char *str, int type, u_int64_t count,
          nptr->flag  = type;
          nptr->count = count;
          nptr->next  = NULL;
-         htab[hash(str)] = nptr;
+         htab[hval]  = nptr;
          if (type!=OBJ_GRP) (*ctr)++;
       }
    }
    else
    {
-      /* hashed */
+      /* hashed (SPEEDUP) */
+      if (l_rnode!=NULL && strcmp(l_rnode->string,str)==0) cptr=l_rnode;
+
       while (cptr != NULL)
       {
          if (strcmp(cptr->string,str)==0)
@@ -498,8 +529,8 @@ int put_rnode(char *str, int type, u_int64_t count,
       {
          nptr->flag  = type;
          nptr->count = count;
-         nptr->next  = htab[hash(str)];
-         htab[hash(str)]=nptr;
+         nptr->next  = htab[hval];
+         htab[hval]  = nptr;
          if (type!=OBJ_GRP) (*ctr)++;
       }
    }
@@ -508,6 +539,7 @@ int put_rnode(char *str, int type, u_int64_t count,
       if (type==OBJ_GRP) nptr->flag=OBJ_GRP;
       else if (isinlist(hidden_refs,nptr->string)!=NULL)
                          nptr->flag=OBJ_HIDE;
+      l_rnode=nptr;
    }
    return nptr==NULL;
 }
@@ -537,6 +569,7 @@ void	del_rlist(RNODEPTR *htab)
          htab[i]=NULL;
       }
    }
+   l_rnode=NULL;
 }
 
 /*********************************************/
@@ -581,11 +614,13 @@ int put_anode(char *str, int type, u_int64_t count,
               u_int64_t *ctr, ANODEPTR *htab)
 {
    ANODEPTR cptr,nptr;
+   uint32_t  hval;
 
    if (str[0]=='-') return 0;     /* skip bad user agents */
 
+   hval = hash(str);
    /* check if hashed */
-   if ( (cptr = htab[hash(str)]) == NULL)
+   if ( (cptr = htab[hval]) == NULL)
    {
       /* not hashed */
       if ( (nptr=new_anode(str)) != NULL)
@@ -593,7 +628,7 @@ int put_anode(char *str, int type, u_int64_t count,
          nptr->flag = type;
          nptr->count= count;
          nptr->next = NULL;
-         htab[hash(str)] = nptr;
+         htab[hval] = nptr;
          if (type!=OBJ_GRP) (*ctr)++;
       }
    }
@@ -618,8 +653,8 @@ int put_anode(char *str, int type, u_int64_t count,
       {
          nptr->flag  = type;
          nptr->count = count;
-         nptr->next  = htab[hash(str)];
-         htab[hash(str)]=nptr;
+         nptr->next  = htab[hval];
+         htab[hval]  = nptr;
          if (type!=OBJ_GRP) (*ctr)++;
       }
    }
@@ -696,18 +731,20 @@ SNODEPTR new_snode(char *str)
 int put_snode(char *str, u_int64_t count, SNODEPTR *htab)
 {
    SNODEPTR cptr,nptr;
+   uint32_t  hval;
 
    if (str[0]==0 || str[0]==' ') return 0;     /* skip bad search strs */
 
+   hval=hash(str);
    /* check if hashed */
-   if ( (cptr = htab[hash(str)]) == NULL)
+   if ( (cptr = htab[hval]) == NULL)
    {
       /* not hashed */
       if ( (nptr=new_snode(str)) != NULL)
       {
          nptr->count = count;
          nptr->next = NULL;
-         htab[hash(str)] = nptr;
+         htab[hval] = nptr;
       }
    }
    else
@@ -727,8 +764,8 @@ int put_snode(char *str, u_int64_t count, SNODEPTR *htab)
       if ( (nptr = new_snode(str)) != NULL)
       {
          nptr->count = count;
-         nptr->next  = htab[hash(str)];
-         htab[hash(str)]=nptr;
+         nptr->next  = htab[hval];
+         htab[hval]  = nptr;
       }
    }
    return nptr==NULL;
@@ -810,11 +847,13 @@ int put_inode( char      *str,  /* ident str */
                INODEPTR  *htab) /* hashtable */
 {
    INODEPTR cptr,nptr;
+   uint32_t  hval;
 
    if ((str[0]=='-') || (str[0]==0)) return 0;  /* skip if no username */
 
+   hval = hash(str);
    /* check if hashed */
-   if ( (cptr = htab[hash(str)]) == NULL)
+   if ( (cptr = htab[hval]) == NULL)
    {
       /* not hashed */
       if ( (nptr=new_inode(str)) != NULL)
@@ -824,7 +863,7 @@ int put_inode( char      *str,  /* ident str */
          nptr->files = file;
          nptr->xfer  = xfer;
          nptr->next  = NULL;
-         htab[hash(str)] = nptr;
+         htab[hval] = nptr;
          if (type!=OBJ_GRP) (*ctr)++;
 
          if (visit)
@@ -871,8 +910,8 @@ int put_inode( char      *str,  /* ident str */
          nptr->count = count;
          nptr->files = file;
          nptr->xfer  = xfer;
-         nptr->next  = htab[hash(str)];
-         htab[hash(str)]=nptr;
+         nptr->next  = htab[hval];
+         htab[hval]  = nptr;
          if (type!=OBJ_GRP) (*ctr)++;
 
          if (visit)
@@ -970,11 +1009,13 @@ DNODEPTR new_dnode(char *str)
 int put_dnode(char *str, void *addr, int len, DNODEPTR *htab)
 {
    DNODEPTR cptr,nptr;
+   uint32_t  hval;
 
    if (str[0]==0 || str[0]==' ') return 0;     /* skip bad hostnames */
 
+   hval = hash(str);
    /* check if hashed */
-   if ( (cptr = htab[hash(str)]) == NULL)
+   if ( (cptr = htab[hval]) == NULL)
    {
       /* not hashed */
       if ( (nptr=new_dnode(str)) != NULL)
@@ -983,7 +1024,7 @@ int put_dnode(char *str, void *addr, int len, DNODEPTR *htab)
             else   memset(&nptr->addr, 0, sizeof(struct sockaddr_storage));
          nptr->addrlen = len;
          nptr->next = NULL;
-         htab[hash(str)] = nptr;
+         htab[hval] = nptr;
       }
    }
    else
@@ -1000,8 +1041,8 @@ int put_dnode(char *str, void *addr, int len, DNODEPTR *htab)
          if (addr) memcpy(&nptr->addr, addr, len);
             else   memset(&nptr->addr, 0, sizeof(struct sockaddr_storage));
          nptr->addrlen = len;
-         nptr->next  = htab[hash(str)];
-         htab[hash(str)]=nptr;
+         nptr->next  = htab[hval];
+         htab[hval]  = nptr;
       }
    }
    return nptr==NULL;
@@ -1036,18 +1077,6 @@ void	del_dlist(DNODEPTR *htab)
 
 #endif /* USE_DNS */
 
-/*********************************************/
-/* HASH - return hash value for string       */
-/*********************************************/
-
-u_int64_t hash(char *str)
-{
-   u_int64_t hashval;
-   for (hashval = 0; *str != '\0'; str++)
-      hashval = *str + 31 * hashval;
-   return hashval % MAXHASH;
-}
-
 /*********************************************/
 /* FIND_URL - Find URL in hash table         */
 /*********************************************/
@@ -1167,3 +1196,84 @@ u_int64_t tot_visit(HNODEPTR *list)
    }
    return tot;
 }
+
+#ifdef USE_OLDHASH
+/*********************************************/
+/* HASH - return hash value for string       */
+/*********************************************/
+
+uint32_t hash(char *str)
+{
+   uint32_t  hashval=0;
+
+   for (hashval = 0; *str != '\0'; str++)
+      hashval = *str + (hashval << 5) - hashval;
+
+   return hashval % MAXHASH;
+}
+
+#else /* USE_OLDHASH */
+/*********************************************/
+/* HASH (SuperFastHash by Paul Hsieh)        */
+/*********************************************/
+
+#undef get16bits
+#if (defined(__GNUC__) && defined(__i386__)) || defined(__WATCOMC__) \
+  || defined(_MSC_VER) || defined (__BORLANDC__) || defined (__TURBOC__)
+#define get16bits(d) (*((const uint16_t *) (d)))
+#endif
+
+#if !defined (get16bits)
+#define get16bits(d) ((((uint32_t)(((const uint8_t *)(d))[1])) << 8)\
+                       +(uint32_t)(((const uint8_t *)(d))[0]) )
+#endif
+
+uint32_t hash(char *str)
+{
+   int len=strlen(str);
+   uint32_t hash = len, tmp;
+   int rem;
+
+   if (len <= 0 || str == NULL) return 0;
+
+   rem = len & 3;
+   len >>= 2;
+
+   /* Main loop */
+   for (;len > 0; len--)
+   {
+      hash  += get16bits (str);
+      tmp    = (get16bits (str+2) << 11) ^ hash;
+      hash   = (hash << 16) ^ tmp;
+      str   += 2*sizeof (uint16_t);
+      hash  += hash >> 11;
+   }
+
+   /* Handle end cases */
+   switch (rem)
+   {
+      case 3: hash += get16bits (str);
+              hash ^= hash << 16;
+              hash ^= str[sizeof (uint16_t)] << 18;
+              hash += hash >> 11;
+              break;
+      case 2: hash += get16bits (str);
+              hash ^= hash << 11;
+              hash += hash >> 17;
+              break;
+      case 1: hash += *str;
+              hash ^= hash << 10;
+              hash += hash >> 1;
+   }
+
+   /* Force "avalanching" of final 127 bits */
+   hash ^= hash << 3;
+   hash += hash >> 5;
+   hash ^= hash << 4;
+   hash += hash >> 17;
+   hash ^= hash << 25;
+   hash += hash >> 6;
+
+   return hash % MAXHASH;
+}
+#endif /* USE_OLDHASH */
