@@ -1,7 +1,7 @@
 /*
     webalizer - a web server log analysis program
 
-    Copyright (C) 1997-2008  Bradford L. Barrett
+    Copyright (C) 1997-2010  Bradford L. Barrett
 
     This program is free software; you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published by
@@ -556,6 +556,21 @@ int parse_record_w3c(char *buffer)
    memset(&gm_time, 0, sizeof(struct tm));
    size = strlen(buffer);                 /* get length of buffer        */
    eob = buffer + size;                   /* calculate end of buffer     */
+
+   /* remove line end markers, reduce eob accordingly */
+   cp = eob;
+   while(cp>buffer)
+   {
+      cp--;
+      if (*cp == '\r' || *cp=='\n')
+      {
+         *cp = '\0';
+         eob--;
+      }
+      else
+         break;
+   }
+
    fmt_logrec(buffer);                    /* seperate fields with \0's   */
 
    cp = buffer;
@@ -575,7 +590,7 @@ int parse_record_w3c(char *buffer)
          while (*cp) cp++;
          cp++;
          index = 1;
-         while ((cp < eob) && (*cp != '\r') && (*cp != '\n'))
+         while (cp < eob)
          {
             /* Set the field index */
             if (!strcmp(cp, "date"))           field_index.date     = index;
@@ -607,7 +622,7 @@ int parse_record_w3c(char *buffer)
    /* Check if the number of entries in this line are conform to the
       format specified in the header */
    index = 1;
-   while ((cp < eob) && (*cp != '\r') && (*cp != '\n'))
+   while (cp < eob)
    {
       while (*cp) cp++;
       cp++;
@@ -621,7 +636,7 @@ int parse_record_w3c(char *buffer)
    /* Reset the field pointers and begin parsing the data line */
    memset(&fields, 0, sizeof(struct fields_struct));
    index = 1;
-   while ((cp < eob) && (*cp != '\r') && (*cp != '\n'))
+   while (cp < eob)
    {
       /* Set the field pointers */
       if (index == field_index.date)     fields.date      = cp;
